import pytest
from starlette.testclient import TestClient

from duf.server.app import app
from duf.server.env import env


@pytest.fixture
def client(tmp_path):
    env.reset(root=tmp_path)
    yield TestClient(app)



import uuid
import json
import getpass
import datetime
import subprocess
from pathlib import Path

import jwt
import click
from fans.logger import get_logger

from .api import api
from .cons import paths


logger = get_logger(__name__)


@click.group()
def auth():
    """Authorization"""
    pass


@auth.command()
@click.option('-e', '--expire-seconds', type=int, default=3600)
def login(expire_seconds):
    """Login and save token"""
    username = input('Username: ')
    password = getpass.getpass()
    data = api.post('https://auth.fans656.me/api/login', {
        'username': username,
        'password': password,
        'expire_seconds': expire_seconds,
    })
    if data:
        paths.token.ensure_parent()
        with paths.token.open('w') as f:
            f.write(data['token'])
        click.echo(f'Login succeed (token written in {paths.token})')
    else:
        exit(1)


@auth.command()
def status():
    """Show login status"""
    if paths.token.exists():
        if not paths.public_key.exists():
            res = api.get('https://auth.fans656.me/public-key', raw=True)
            if res.status_code == 200:
                with paths.public_key.open('w') as f:
                    f.write(res.text)
            else:
                click.echo(click.style('failed to get public key', fg='red'))
                exit(1)
        with paths.public_key.open() as f:
            public_key = f.read()
        with paths.token.open() as f:
            token = f.read()
        data = jwt.decode(
            token,
            public_key,
            algorithms=['RS256'],
            options={"verify_signature": False, "verify_exp": False},
        )
        click.echo('-' * 80 + 'public key')
        click.echo(public_key)
        click.echo('-' * 80 + 'token')
        click.echo(token)
        click.echo('-' * 80 + 'decoded')
        click.echo(json.dumps(data, indent=2))
        if 'exp' in data:
            exp = datetime.datetime.fromtimestamp(data['exp'])
            exp_str = exp.strftime('%Y-%m-%d %H:%M:%S')
            message = f'Expiration {exp_str}'
            if exp <= datetime.datetime.now():
                message = click.style(message, fg='red')
            click.echo(message)
    else:
        click.echo(click.style('not logged in', fg='red'))



import click

from .log import log
from .init import init
from .host import host
from .auth import auth
from .deploy import deploy
from .utils import execute


@click.group
def cli():
    pass


cli.add_command(init)
cli.add_command(host)
cli.add_command(auth)
cli.add_command(deploy)
cli.add_command(log)


@cli.command
def publish():
    execute('uv build')
    execute('uv publish')
    execute('rm -r dist *.egg-info')



import subprocess

from fans.logger import get_logger


logger = get_logger(__name__)


def execute(cmd: str|list[str]):
    if isinstance(cmd, list):
        cmd = ' '.join(cmd)
    logger.info(f'[CMD] {cmd}')
    proc = subprocess.run(cmd, shell=True)
    if not proc.returncode == 0:
        logger.error(f'failed to execute command: {cmd}')
        exit(1)



import subprocess
from pathlib import Path

import click
from fans.logger import get_logger

from .api import api


logger = get_logger(__name__)


@click.command()
def log():
    """Show running instance log"""
    pass
    #os.system('')



from fans.path import Path, make_paths


paths = make_paths(
    Path.home(), [
        '.duf', [
            'auth', [
                'token', {'token'},
                'public_key', {'public_key'},
            ],
        ],
    ],
)



from fans.path import Path, make_paths


_server_dir = Path(__file__).parent
_duf_dir = _server_dir.parent
_proj_root_dir = _duf_dir.parent


class Env:

    def __init__(self):
        self.reset()

    def reset(self, root = ''):
        self.paths = make_paths([
            Path(root), [
                'data', [
                    'hosts.json', {'hosts'},
                ],
            ],
            _proj_root_dir, [
                'scripts', [
                    'init-duf.sh', {'init_duf_sh'},
                ],
            ],
        ])


env = Env()



import json
import subprocess
from pathlib import Path

import click
from fans.logger import get_logger

from .api import api


logger = get_logger(__name__)


@click.command()
@click.option('-S', '--no-sync', is_flag=True)
def deploy(no_sync):
    """Deploy current project"""
    cwd = Path().absolute()
    proj_name = cwd.name
    host = get_target_host()

    #execute(f"ssh root@{host['ip']} 'curl -s https://duf.fans656.me/init-duf.sh | bash'")

    if not no_sync:
        execute([
            f"rsync -rav",

            f"--exclude '**/__pycache__/'",
            f"--exclude '**/node_modules/'",

            f"--include '{proj_name}/'",
            f"--include '{proj_name}/**'",

            f"--include 'frontend/'",
            f"--include 'frontend/dist/'",
            f"--include 'frontend/dist/**'",

            f"--include 'scripts/'",
            f"--include 'scripts/**'",

            f"--include 'pyproject.toml'",
            f"--include 'uv.lock'",
            f"--include 'serve.sh'",

            f"--exclude '*'",

            f"./ root@{host['ip']}:/root/{proj_name}",
        ])

    execute(
        f"ssh root@{host['ip']} "
        f"uv --directory /root/duf run -m "
        f"duf.cli.on_host_deploy --target /root/{proj_name}"
    )
    #execute(f"ssh root@{host['ip']} supervisorctl restart {proj_name}")


def get_target_host():
    cache_json = Path('./.duf-cache.json')
    if not cache_json.exists():
        with cache_json.open('w') as f:
            json.dump({}, f)

    with cache_json.open() as f:
        data = json.load(f)
        if 'host' in data:
            return data['host']

    logger.info('getting target host to deploy...')
    data = api.post('/api/host/ls')
    if data:
        for host in data['hosts']:
            if host.get('name') == 'default':
                logger.info(f'got target host {host}')

                data['host'] = host
                with cache_json.open('w') as f:
                    json.dump(data, f, indent=2)

                return host
    logger.error('failed to get target host to deploy')
    exit(1)


def execute(cmd: str|list[str]):
    if isinstance(cmd, list):
        cmd = ' '.join(cmd)
    logger.info(f'[CMD] {cmd}')
    proc = subprocess.run(cmd, shell=True)
    if not proc.returncode == 0:
        logger.error(f'failed to execute command: {cmd}')
        exit(1)



import os
import json

import click
import requests

from .cons import paths


class API:

    def __init__(self):
        self.origin = 'https://duf.fans656.me'

    def get(self, *args, **kwargs):
        return self.request('GET', *args, **kwargs)

    def post(self, *args, **kwargs):
        return self.request('POST', *args, **kwargs)

    def request(self, method, path, req: dict = {}, raw: bool = False):
        if path.startswith('http'):
            url = path
        else:
            url = f'{self.origin}{path}'

        cookies = {}
        if paths.token.exists():
            with paths.token.open() as f:
                cookies['token'] = f.read()

        if method == 'GET':
            res = requests.get(
                url,
                params=req,
                cookies=cookies,
            )
        else:
            res = requests.post(
                url,
                json=req,
                cookies=cookies,
            )

        if res.status_code == 200:
            if raw:
                return res
            else:
                return res.json()
        else:
            try:
                click.echo(json.dumps(res.json(), indent=2))
            except:
                click.echo(res.text)
            click.echo(click.style(f'ERROR: {res.status_code}', fg='red'))


api = API()



import click

from .utils import execute


@click.group()
def init():
    """Initialize project"""
    pass


@init.command()
def frontend():
    """Initialize frontend"""
    execute('npm create vite@latest -y frontend -- --template react')



import sys
import uuid
from typing import Optional

from fastapi import FastAPI, Body, HTTPException
from fastapi.middleware.cors import CORSMiddleware
from starlette.responses import FileResponse
from fansauth import auth

from .env import env


app = FastAPI()
if 'pytest' not in sys.argv[0] and '-d' not in sys.argv:
    app = auth(app, login='api')


app.add_middleware(
    CORSMiddleware,
    allow_origins = [
        'https://fans656.me',
        'http://localhost',
        'https://localhost.fans656.me',
    ],
    allow_credentials=True,
    allow_methods=['*'],
    allow_headers=['*'],
)


@app.get('/init-duf.sh')
def init_duf_sh():
    return FileResponse(env.paths.init_duf_sh)


@app.get('/api/host/list')
def host_list():
    return list(Hosts())


@app.post('/api/host/ls')
def host_ls():
    return {'hosts': list(Hosts())}


@app.post('/api/host/info')
def host_info(req: dict):
    host = Hosts().find(req['target'])
    if host:
        return host
    else:
        raise HTTPException(404)


@app.post('/api/host/add')
def host_add(host: dict):
    host_id = Hosts().add(host)
    return {'id': host_id}


@app.post('/api/host/edit')
def host_edit(update: dict):
    host = Hosts().update(update)
    if host:
        return host
    else:
        raise HTTPException(404)


@app.post('/api/host/remove')
def host_remove(req: dict):
    host = Hosts().remove(req['target'])
    if host:
        return host
    else:
        raise HTTPException(404)


class Hosts:

    def __init__(self):
        self.meta = env.paths.hosts.as_meta(default={
            'hosts': [],
        })

    def add(self, host: dict) -> Optional[str]:
        host.setdefault('id', uuid.uuid4().hex)
        self.meta['hosts'].append(host)
        self.meta.save()
        return host['id']

    def update(self, update: dict):
        host = self.get(update['id'])
        if host:
            host.update(update)
            self.meta.save()
            return host
        else:
            return None

    def find(self, target: str):
        for host in self:
            if is_target(host, target):
                return host

    def get(self, host_id: str):
        return next((d for d in self if d['id'] == host_id), None)

    def remove(self, target: str):
        host = self.find(target)
        if host:
            self.meta['hosts'] = [d for d in self if not is_target(d, target)]
            self.meta.save()
            return host

    def __iter__(self):
        for host in self.meta['hosts']:
            yield host

    def __contains__(self, host_id: str):
        return any(host_id == host['id'] for host in self)


def is_target(host: dict, target: str):
    if host['id'] == target:
        return True
    if host['name'] == target:
        return True
    return False



class Test_init_duf:

    def test_get_init_duf_sh(self, client):
        res = client.get('/init-duf.sh')
        assert res.status_code == 200
        assert res.text


class Test_host:

    def test_add_and_ls(self, client):
        res = client.post('/api/host/add', json={
            'name': 'foo',
            'ip': '1.2.3.4',
        })
        assert res.status_code == 200

        res = client.post('/api/host/ls')
        assert res.status_code == 200
        hosts = res.json()['hosts']
        assert len(hosts) == 1
        host = hosts[0]
        assert host['name'] == 'foo'
        assert host['ip'] == '1.2.3.4'

    def test_edit_and_info_and(self, client):
        res = client.post('/api/host/add', json={
            'name': 'foo',
            'ip': '1.2.3.4',
        })
        host_id = res.json()['id']

        res = client.post('/api/host/edit', json={
            'id': host_id,
            'name': 'bar',
        })
        assert res.json() == {'id': host_id, 'name': 'bar', 'ip': '1.2.3.4'}

        res = client.post('/api/host/info', json={'target': 'bar'})
        assert res.json() == {'id': host_id, 'name': 'bar', 'ip': '1.2.3.4'}

    def test_edit_not_found(self, client):
        res = client.post('/api/host/edit', json={
            'id': 'foo',
        })
        assert res.status_code == 404

    def test_remove(self, client):
        res = client.post('/api/host/add', json={'name': 'foo'})
        host_id = res.json()['id']

        client.post('/api/host/add', json={'name': 'bar'})

        client.post('/api/host/remove', json={'target': host_id})
        client.post('/api/host/remove', json={'target': 'bar'})

        res = client.post('/api/host/ls')
        assert not res.json()['hosts']



import os
import json
import uuid
import subprocess
from pathlib import Path

import yaml
import click

from .api import api


@click.group()
def host():
    """Host management"""
    pass


@host.command()
def ls():
    """List available hosts"""
    data = api.post('/api/host/ls')
    if data:
        for host in data['hosts']:
            click.echo(f'{host}')


@host.command()
@click.argument('name')
def add(name: str):
    """Add new host"""
    data = api.post('/api/host/add', {'name': name})
    if data:
        click.echo(f'Added {data}')


@host.command()
@click.argument('target')
def edit(target: str):
    """Edit host info"""
    old_data = api.post('/api/host/info', {'target': target})
    if not old_data:
        return
    fpath = Path('DUF_TARGET_EDITING')
    try:
        with fpath.open('w') as f:
            yaml.dump(old_data, f)

        editor = os.getenv('EDITOR', 'vi')
        subprocess.run(f'{editor} {fpath}', shell=True)

        with fpath.open() as f:
            new_data = yaml.safe_load(f)

        if old_data == new_data:
            click.echo('Nothing changed')
        else:
            api.post('/api/host/edit', new_data)
    finally:
        if fpath.exists():
            fpath.unlink()


@host.command()
@click.argument('target')
def info(target: str):
    """Show host info"""
    host = api.post('/api/host/info', {'target': target})
    if host:
        click.echo(json.dumps(host, indent=2))


@host.command()
@click.argument('target')
def rm(target: str):
    """Remove host"""
    data = api.post('/api/host/remove', {'target': target})
    if data:
        click.echo(f'Removed {data["id"]}')



import os
from pathlib import Path

import click


@click.command
@click.option('-t', '--target')
def cli(target):
    proj_dir = Path(target)
    print(f'project directory: {proj_dir}')
    proj_name = proj_dir.name
    print(f'project name: {proj_name}')
    supervisor_conf_path = Path(f'/etc/supervisor/conf.d/{proj_name}.conf')

    if not supervisor_conf_path.exists():
        with supervisor_conf_path.open('w') as f:
            f.write(supervisor_conf_template.format(
                proj_name=proj_name,
            ))
        print(f'written supervisor conf: {supervisor_conf_path}')

        execute('supervisorctl reread')
        execute('supervisorctl update')

    execute(f'supervisorctl restart {proj_name}')


def execute(cmd):
    print(f'[CMD] {cmd}')
    os.system(cmd)


supervisor_conf_template = '''
[program:{proj_name}]
command=/root/{proj_name}/serve.sh
directory=/root/{proj_name}
autostart=true
autorestart=true
stderr_logfile=/var/log/{proj_name}.err.log
stdout_logfile=/var/log/{proj_name}.out.log
stopasgroup=true
killasgroup=True
'''


if __name__ == '__main__':
    cli()
